/*
* SimCoupe : Copyright (c) Allan Skillman 1996.
*
* enhanced by Copyright (c) Aley Keprt 1999
*
* File fdi.c - VL 1772 disk controller emulation for SimCoupe
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
typedef long blkcnt_t;
#include <sys/stat.h>

#ifdef linux
#include <sys/file.h>
#include <sys/ioctl.h>
#include <linux/fd.h>
#endif


#ifdef __DJGPP

 #include "floppy.h"

#endif



#include <unistd.h>

#include "z80.h"
#include "sam.h"
#include "util.h"
#include "fdi.h"

#include "zlib.h"


typedef struct _vl1772_reg {
 uns8 Command;
 uns8 Status;
 uns8 Track;
 uns8 Sector;
 uns8 Data;
} vl1772_reg;


typedef struct _DiskAttr {
 int ImageType;
 uns8 sid;
 uns8 trk;
 uns8 sec;
 uns8 SectorDiv64;
} DiskAttr;


typedef struct _Aside {
 char  *FileName;
 FILE  *FPtr;

 uns8 *data; //contents of a disk
 int changed;

 DiskAttr Dattr;
 vl1772_reg reg;
 uns8  SectorBuf[MAX_SECTOR_SIZE];
 uns8  *Bufp;
 int  ByteCount;
} ASide;


#define D_DISK(x) (((x) & 0x10) >> 4)
#define D_SIDE(x) (((x) & 0x04) >> 2)
#define DISK(x)  TheDisks[D_DISK((x) & 0xff)][D_SIDE((x) & 0xff)]

enum {COMMAND = 0, STATUS = 0, TRACK, SECTOR, DATA};


static ASide TheDisks[2][MAX_NO_SIDES];


int ReadSector (uns8 Side_no, ASide *Dside);
int WriteSector(uns8 Side_no, ASide *Dside);


void fdi_out(uns16 port, uns8 byte) {
 int Sdir = 1;
 
 switch (port & 0x03) {
  
 case COMMAND:
  /*                        printf("Command : %x\n",(byte & 0xf0)); */
  DISK(port).reg.Command = byte & 0xf0;
  
  switch (byte & 0xf0) {
   /* Type I commands */
  case RESTORE:
   DISK(port).reg.Track = 0;
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP | LOST_DATA;
   break;
  case SEEK:
   DISK(port).reg.Track = DISK(port).reg.Data;
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   if (!DISK(port).reg.Track) 
    DISK(port).reg.Status |= LOST_DATA;
   break;
  case STEP_NUPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   break;
  case STEP_UPD:
   DISK(port).reg.Track += Sdir;
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   if (!DISK(port).reg.Track) 
    DISK(port).reg.Status |= LOST_DATA;
   break;
  case STEP_IN_NUPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   break;
  case STEP_IN_UPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   DISK(port).reg.Track++;
   break;
  case STEP_OUT_NUPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   break;
  case STEP_OUT_UPD:
   DISK(port).reg.Status = DRQ | MOTOR_ON | SPIN_UP;
   if (!DISK(port).reg.Track--) 
    DISK(port).reg.Status |= LOST_DATA;
   break;
   /* Type II Commands */
  case READ_1SECTOR:
  case READ_MSECTOR:
   DISK(port).reg.Status = BUSY | DRQ;
   if (ReadSector(D_SIDE(port),&DISK(port))) {
    DISK(port).reg.Status = RECORD_NOT_FOUND;
   }
   break;
  case WRITE_1SECTOR:
  case WRITE_MSECTOR:
   DISK(port).reg.Status = BUSY | DRQ;
   DISK(port).Bufp = DISK(port).SectorBuf;
   DISK(port).ByteCount = 0;
   break;
   /* Type III Commands */
  case READ_ADDRESS:
  case READ_TRACK:
  case WRITE_TRACK:
   break;
   /* Type IV Commands */
  case FORCE_INTERRUPT:
   DISK(port).reg.Status &= ~BUSY;
   DISK(port).reg.Command = 0;
   break;
   
  default:
   printf("Unknown Disk Command %x\n",byte & 0xf0);
   break;
  } /* End switch */
  break;
  
  case TRACK:
   DISK(port).reg.Track = byte;
   break;
  case SECTOR:
   DISK(port).reg.Sector = byte;
   break;
  case DATA:
   if (DISK(port).reg.Status & BUSY) {
    if (DISK(port).reg.Command == WRITE_1SECTOR ||
     DISK(port).reg.Command == WRITE_MSECTOR) {
     *DISK(port).Bufp++ = byte;
     /*                                printf("ByteCount = %d\n",DISK(port).ByteCount);*/
     if (++DISK(port).ByteCount == (64*DISK(port).Dattr.SectorDiv64) ) {
      if (WriteSector(D_SIDE(port),&DISK(port))) {
       DISK(port).reg.Status = RECORD_NOT_FOUND;
       break;
      }
      if (DISK(port).reg.Command == READ_MSECTOR &&
       DISK(port).reg.Sector++ != DISK(port).Dattr.sec) {
       DISK(port).Bufp = DISK(port).SectorBuf;
       DISK(port).ByteCount = 0;
      } else {
       DISK(port).reg.Status &= ~(BUSY | DRQ);
       DISK(port).reg.Command = 0;
      }     
     }
    } 
   } 
   DISK(port).reg.Data = byte;
   break;
 } /* End Switch */
}


uns8 fdi_in(uns16 port) {
 uns8 byte;
 
 static uns8 DRQ_count = 7;
 
 switch (port & 0x03) {
  
 case STATUS :
  /*                        printf("Status : %x\n", DISK(port).reg.Status); */
  byte = DISK(port).reg.Status;
  if ((!DISK(port).reg.Command) && !DRQ_count--) {
   DISK(port).reg.Status ^= DRQ;
   DRQ_count = 7;
  }
  break;
 case TRACK:
  byte = DISK(port).reg.Track;
  break;
 case SECTOR:
  byte = DISK(port).reg.Sector;
  break;
 case DATA:
  /*                        printf("Data : %x\n",DISK(port).ByteCount); */
  if (DISK(port).reg.Status & BUSY) {
   if (DISK(port).reg.Command == READ_1SECTOR ||
    DISK(port).reg.Command == READ_MSECTOR) {
    DISK(port).reg.Data = *DISK(port).Bufp++;
    if (++DISK(port).ByteCount == (64*DISK(port).Dattr.SectorDiv64) ) {
     if (DISK(port).reg.Command == READ_MSECTOR &&
      DISK(port).reg.Sector++ != DISK(port).Dattr.sec) {
      if (ReadSector(D_SIDE(port),&DISK(port))) {
       DISK(port).reg.Status = RECORD_NOT_FOUND;
       break;
      }
     } else {
      DISK(port).reg.Status &= ~(BUSY | DRQ);
      DISK(port).reg.Command = 0;
     }     
    }
   } 
  } 
  byte = DISK(port).reg.Data;
  break;
 } /* End Switch */
 
 return(byte);
}


uns8* load_SAD_image(char *fn,DiskAttr *attr) {
/*
  Check whether the disk image file fn is a SAD image
  Fill in the disk attributes from the file

  Loads disk into memeory buffer, returns pointer to data
*/

 #define SAD_ID_LENGTH (sizeof(SAD_FORMAT_ID)-1)

 gzFile *gzf;
 char id[SAD_ID_LENGTH];
 long dsize;
 uns8 *data=NULL;

 gzf=gzopen(fn,"rb");
 if(!gzf) return NULL;

 if(gzread(gzf,id,SAD_ID_LENGTH)==SAD_ID_LENGTH) {

  if(!memcmp(id,SAD_FORMAT_ID,SAD_ID_LENGTH)) {
   attr->ImageType=SAD_IMAGE;
   attr->sid=gzgetc(gzf);
   attr->trk=gzgetc(gzf);
   attr->sec=gzgetc(gzf);
   attr->SectorDiv64=gzgetc(gzf);

   /* load data */
   dsize=attr->sid*attr->trk*attr->sec*attr->SectorDiv64*64;
   data=malloc(dsize);
   if(data) {
    if(gzread(gzf,data,dsize)==dsize) {
     gzclose(gzf);
     return data;
    }
   }
  }
 }

 if(data) free(data);
 gzclose(gzf);
 return 0;
}


void insert_disk(char *fn, int disk_no) {
 
 struct stat stbuff;
 DiskAttr tmpAttr;
 FILE *TheFPtr;
 uns8 *data;

 disk_no--;
 
 /* Try to work out the type of disk image inserted */
 
 if (!stat(fn, &stbuff)) {
  
  if ((stbuff.st_mode & S_IFMT) == S_IFREG) {
   /* Regular file */
   
   /* Check for a SAD image */
   if((data=load_SAD_image(fn,&tmpAttr))!=NULL) {

    simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): SAD type image : %s",fn);

   } else {

    switch (stbuff.st_size) {
    case SAM_FORMAT_SIZE:
     simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): SAM type image : %s",fn);
     tmpAttr.ImageType = SAM_IMAGE;
     tmpAttr.sid = 2;
     tmpAttr.trk = 80;
     tmpAttr.sec = 10;
     tmpAttr.SectorDiv64 = 8;
     break;
    case MSDOS_FORMAT_SIZE:
     simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): MSDOS/CPM type image : %s",fn);
     tmpAttr.ImageType = MSDOS_IMAGE;
     tmpAttr.sid = 2;
     tmpAttr.trk = 80;
     tmpAttr.sec = 9;
     tmpAttr.SectorDiv64 = 8;
     break;
    default:
     simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Unknown image type : %s",fn);
     return;
     break;
    } /* End switch */
   }
  }
  else if(((stbuff.st_mode & S_IFMT) == S_IFBLK)

#ifdef linux
   && ( stbuff.st_rdev/256 == 2))  {
   /* For Linux Major device 2 is a floppy drive */
   int fd;
   struct floppy_struct fstruct;
   
   fd = open(fn, O_RDONLY, 0);
   if (!ioctl(fd, FDGETPRM, &fstruct)) {
    tmpAttr.sid   = fstruct.head;
    tmpAttr.trk  = fstruct.track;
    tmpAttr.sec = fstruct.sect;
    close(fd);
   }
   else {
    simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Cannot get floppy parameters : %s",fn);
    simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Assuming Sam Disk");
    tmpAttr.sid = 2;
    tmpAttr.trk = 80;
    tmpAttr.sec = 10;
    close(fd);
   } 
  }
  
#else

  ) {
   simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Direct access to floppy disk only"
    " available on Linux systems");
   return;
  }
#endif 

  else {
   simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Wrong file type : %s",fn);
   return;   
  } if(tmpAttr.sec > 10) {
   simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): Cannot use disks with more than 10"
    "sectors/track\n : %s",fn);
   return;
  }

  /* Pick up the File Pointer of the image */
  if(tmpAttr.ImageType!=SAD_IMAGE) TheFPtr=fopen(fn,"rb+");

 } else if (!strcmp(fn,FDI_FLOPPY_1) ) {

  /* Use the PC floppy drive 1 */ 
  simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): Real floppy drive (1)");
  tmpAttr.ImageType = REAL_FD;
  tmpAttr.sid = 0;  /* sid stores PC drive No */
  tmpAttr.trk = 80;
  tmpAttr.sec = 10;
  tmpAttr.SectorDiv64 = 8;
  
  TheFPtr = NULL;
#ifdef __DJGPP
  FloppyInit(0);
#endif

 } else if (!strcmp(fn,FDI_FLOPPY_2) ) {

  /* Use the PC floppy drive 2 */ 
  simcoupe_mesg(SIMCOUPE_INFO,"insert_disk(): Real floppy drive (2)");
  tmpAttr.ImageType = REAL_FD;
  tmpAttr.sid = 1;  /* sid stores PC drive No */
  tmpAttr.trk = 80;
  tmpAttr.sec = 10;
  tmpAttr.SectorDiv64 = 8;
  
  TheFPtr = NULL;
#ifdef __DJGPP
  FloppyInit(1);
#endif

 } else {
  simcoupe_mesg(SIMCOUPE_WARN,"insert_disk(): cannot stat %s",fn);  
  return;
 }

 TheDisks[disk_no][0].FileName = (char *)malloc(strlen(fn)+1);
 strcpy(TheDisks[disk_no][0].FileName, fn);
 TheDisks[disk_no][1].FileName = TheDisks[disk_no][0].FileName;

 {
 int i;
 ASide *side=TheDisks[disk_no];
 for(i=0;i<2;i++) {
  side[i].FPtr = TheFPtr;
  side[i].Dattr = tmpAttr;
  side[i].data=data;
  side[i].changed=0;
 }
 }
}


void eject_disk(int disk_no) {

 struct _Aside *side=TheDisks[--disk_no];

 //perform delayed save, if necessary
 if(side->Dattr.ImageType==SAD_IMAGE &&
  (side[0].changed || side[1].changed)) {
  gzFile *gzf;
  struct _DiskAttr *attr=&side->Dattr;
  long dsize=attr->sid*attr->trk*attr->sec*attr->SectorDiv64*64;

  gzf=gzopen(side->FileName,"wb");
  if(gzf) {
   gzwrite(gzf,SAD_FORMAT_ID,SAD_ID_LENGTH);
   gzputc(gzf,attr->sid);
   gzputc(gzf,attr->trk);
   gzputc(gzf,attr->sec);
   gzputc(gzf,attr->SectorDiv64);

   /* save data */
   gzwrite(gzf,side->data,dsize);
   gzclose(gzf);
  }
 }

 free(TheDisks[disk_no][0].FileName);
 side[0].FileName = NULL;
 side[1].FileName = NULL;
 
 if(TheDisks[disk_no][0].FPtr) fclose(TheDisks[disk_no][0].FPtr);
 
 side[0].FPtr = NULL;
 side[1].FPtr = NULL;

 free(TheDisks[disk_no][0].data);
 side[0].data=NULL;
 side[1].data=NULL;
}


int ReadSector(uns8 Side_no, ASide *Dside) {
 long offset;
 int i;
 
 if (Dside->FPtr || Dside->FileName) {

  switch(Dside->Dattr.ImageType) {

  case REAL_FD:
#ifdef __DJGPP
   FloppyRead( Dside->Dattr.sid,
    Side_no, Dside->reg.Track,
    Dside->reg.Sector,
    Dside->SectorBuf);
   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
#endif
   return 0; 
   
  case SAD_IMAGE:
   offset =// SAD_HEADER +
    (64*Dside->Dattr.SectorDiv64) * 
    (Dside->Dattr.sec * (Dside->Dattr.trk*Side_no + Dside->reg.Track) +
    (Dside->reg.Sector - 1)
    );

   //move data from memory buffer
   memcpy(Dside->SectorBuf,Dside->data+offset,64*Dside->Dattr.SectorDiv64);

   //reset sector counters
   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
   return 0;

  case SAM_IMAGE:
  case MSDOS_IMAGE:
   offset = ((64*Dside->Dattr.SectorDiv64) * Dside->Dattr.sec) * 
    (2*Dside->reg.Track + Side_no) +
    (64*Dside->Dattr.SectorDiv64)  * (Dside->reg.Sector - 1);
   break;

  default:
   return -1;
  }
  
  /*         printf("Side = %d,  Track = %d, Sector = %d, Offset = %d\n", 
  Side_no, Dside->reg.Track, Dside->reg.Sector, offset);
  */

  if (fseek(Dside->FPtr, offset, SEEK_SET) != 0) {
   return -2; /* couldn't seek the required sector */  
  } else {
   if(fread(Dside->SectorBuf, sizeof(uns8), (64*Dside->Dattr.SectorDiv64) , Dside->FPtr) != (64*Dside->Dattr.SectorDiv64) ) {
    return -3; /* couldn't read the required sector */
   } else {
   /*
   for (i = 0; i < 512 ; i++) 
   printf("%x",Dside->SectorBuf[i]);
   printf ("\n");
    */
    Dside -> ByteCount = 0;
    Dside -> Bufp = Dside -> SectorBuf;
    return 0; 
   }
  }
 } else {
  return -1; /* couldn't open image => drive not ready */
 }
}


int WriteSector(uns8 Side_no, ASide *Dside) {
 long offset;

 if(Dside->FPtr || Dside->FileName) {
  if(Dside->Dattr.ImageType == REAL_FD) {
#ifdef __DJGPP
   
   FloppyWrite(Dside->Dattr.sid,
    
    Side_no, Dside->reg.Track,
    
    Dside->reg.Sector,
    
    Dside->SectorBuf);
   
   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
#endif
  } else if (Dside->Dattr.ImageType == SAD_IMAGE) {

   offset =// SAD_HEADER +
    (64*Dside->Dattr.SectorDiv64) * 
    (Dside->Dattr.sec * (Dside->Dattr.trk*Side_no + Dside->reg.Track) +
    (Dside->reg.Sector - 1)
    );

   //move data to the memory buffer
   memcpy(Dside->data+offset,Dside->SectorBuf,64*Dside->Dattr.SectorDiv64);

   //request delayed save
   Dside->changed=1;

   Dside -> ByteCount = 0;
   Dside -> Bufp = Dside -> SectorBuf;
   return 0;

  } else {

   offset = ((64*Dside->Dattr.SectorDiv64) * Dside->Dattr.sec) * 
    (2*Dside->reg.Track + Side_no) +
    (64*Dside->Dattr.SectorDiv64)  * (Dside->reg.Sector - 1);
   
  }
  /*         printf("Side = %d,  Track = %d, Sector = %d, Offset = %d\n", 
  Side_no, Dside->reg.Track, Dside->reg.Sector, offset);*/
  
  if(fseek(Dside->FPtr, offset, SEEK_SET) != 0) {
   return -2; /* couldn't seek the required sector */  
  } else {
   if(fwrite(Dside->SectorBuf, sizeof(uns8), (64*Dside->Dattr.SectorDiv64) , Dside->FPtr) != (64*Dside->Dattr.SectorDiv64) ) {
    return -3; /* couldn't write that sector */
   } else {
    fflush(Dside->FPtr);
    Dside -> ByteCount = 0;
    Dside -> Bufp = Dside -> SectorBuf;
    return 0;
   }
  }
 } else {
  return -1;
 }
}


int create_disk_image(char *ImageFile, int AddSystem) {
 /* Create a blank disk image (.dsk format) and */
 /* add a DOS system (SYSTEM_DOS) if required.  */
 int   iTrack;
 uns8 *TrackBuffer;
 FILE *ImageFp, *SystemFp;
 int   Status = 0;
 
 TrackBuffer = (uns8 *)malloc(5120*sizeof(uns8));
 
 memset(TrackBuffer, 0, 5120*sizeof(uns8));
 
 if (ImageFp=fopen(ImageFile, "wb")) {
  /* First create a blank image */
  for (iTrack=0; iTrack < 160; iTrack++) {
   Status += !(fwrite(TrackBuffer, 512, 10, ImageFp) == 10);
  }
  
  /* Add the system if required */
  if(AddSystem) {
   rewind(ImageFp);
   if(SystemFp=fopen(SYSTEM_DOS, "rb")) {
    /* First copy the directory entry */
    Status+=!(fread (TrackBuffer,256,1,SystemFp)==1);
    Status+=!(fwrite(TrackBuffer,256,1,ImageFp) ==1);
    
    /* Seek to the first data sector */
    Status+=fseek(ImageFp, 0xa000, SEEK_SET);
    Status+=!(fread (TrackBuffer,512,10,SystemFp)==10);
    Status+=!(fwrite(TrackBuffer,512,10,ImageFp) ==10);
    
    Status+=fseek(ImageFp, 0xc800, SEEK_SET);
    Status+=!(fread (TrackBuffer,512,10,SystemFp)==10);
    Status+=!(fwrite(TrackBuffer,512,10,ImageFp) ==10);
    fclose(SystemFp);
   } else {
    Status = 1;    
   }
  }
  fclose(ImageFp);
 }
 else {
  Status = 1;
 }
 
 free(TrackBuffer);
 return Status;
}